int ramses(char *cmd, char *out);
int get_volt_mag(char *bus_name, double *bus_volt);
int get_volt_pha(char *bus_name, double *bus_pha);
int get_line_pow(char *line_name, double *p_orig, double *q_orig, double *p_extr, double *q_extr);
int get_line_cur(char *line_name, double *ix_orig, double *iy_orig, double *ix_extr, double *iy_extr);
double get_sim_time(void);
double get_huge_double(void);
int get_mxprm(void);
int get_comp_prm_names(char *comp_type, char *comp_name, int mxprm, char *names);
int define_SS(int ssID, char *filter1, int num1, char *filter2, int num2,char *filter3, int num3);
int get_SS(int ssID, int mxreclen, char *busNames);
int get_transformer_in_SS(int ssID, int location, int in_service, char *rettype, int mxreclen, char *retString, double *retdp, int *retint, int *elem);
int get_nbbus(void);
int get_bus_name(int i, char *name);
int get_nbsync(void);
int get_sync_name(int i, char *name);
int get_nbinj(void);
int get_inj_name(int i, char *name);
int get_nbdctl(void);
int get_dctl_name(int i, char *name);
int get_nbbra(void);
int get_branch_name(int i, char *name);
int get_nbtwop(void);
int get_twop_name(int i, char *name);
int get_nbshunt(void);
int get_shunt_name(int i, char *name);
int get_nbload(void);
int get_load_name(int i, char *name);
int set_pause_time(double c_pause_time);
int continue_simul(void);
int set_to_backup(void);
int set_end_simul(void);
int get_end_simul(void);
int add_disturb(double disturbtime, char *disturbdesc);
int get_named_prm(char *comp_type, char *comp_name, char *prm_name, double *retd);
int get_named_obs(char *comp_type, char *comp_name, char *obs_name, double *retd);
int initObserv(char *traj_filenm);
int addObserv(char *string);
int finalObserv(void);
int get_last_err_log(char *msg);
int get_Jac(void);
